@@ -1,22 +1,27 @@
 """ A class for interacting with Aussie Broadband APIs """
 
-import inspect
-
 import json
 from time import time
-import sys
+from typing import Any, Dict, List, Optional
 
-from loguru import logger
 import requests
+import requests.sessions
 
-from .const import BASEURL, API_ENDPOINTS, default_headers
-from .exceptions import AuthenticationException, RateLimitException, RecursiveDepth
-from .utils import get_url
+from .baseclass import BaseClass
+from .const import BASEURL, default_headers, PHONE_TYPES
+from .exceptions import RecursiveDepth
+from .types import ServiceTest, AccountTransaction
 
-#pylint: disable=too-many-public-methods,too-many-instance-attributes
-class AussieBB():
+class AussieBB(BaseClass):
     """ A class for interacting with Aussie Broadband APIs """
-    def __init__(self, username: str, password: str, debug: bool=False, **kwargs):
+    def __init__(
+        self,
+        username: str,
+        password: str,
+        debug: bool=False,
+        services_cache_time: int = 28800,
+        session : requests.sessions.Session = None,
+        ):
         """ Setup function
 
             ```
@@ -26,30 +31,20 @@ class AussieBB():
             @param services_cache_time: int
                 - seconds between caching get_services()
                 - defaults to 8 hours
+            @param session : requests.session - session object
             ```
         """
-        self.username = username
-        self.password = password
-
-        self.debug = debug
-        # TODO: set the debug level in loguru when this is set to False
-
-        if not (username and password):
-            raise AuthenticationException("You need to supply both username and password")
-        self.session = requests.Session()
-
-        self.myaussie_cookie = ""
-        self.token_expires = -1
-
-        self.services_cache_time = kwargs.get('services_cache_time', 28800) # defaults to 8 hours
-        self.services_last_update = -1
-        self.services = None
+        super().__init__(username, password, debug, services_cache_time)
+        if session is None:
+            self.session = requests.Session()
+        else:
+            self.session = session
 
     def login(self, depth=0):
         """ Logs into the account and caches the cookie.  """
         if depth>2:
             raise RecursiveDepth("Login recursion depth > 2")
-        logger.debug("Logging in...")
+        self.logger.debug("Logging in...")
 
         url = BASEURL.get('login')
 
@@ -61,39 +56,28 @@ class AussieBB():
 
         response = self.session.post(url,
                                      headers=headers,
-                                     data=json.dumps(payload),
+                                     json=payload,
                                      )
-        if response.status_code == 422:
-            raise AuthenticationException(response.json())
-        if response.status_code == 429:
-            raise RateLimitException(response.json())
-        response.raise_for_status()
 
+        response.raise_for_status()
         jsondata = response.json()
 
-        # we expire a little early just because
-        self.token_expires = time() + jsondata.get('expiresIn') - 50
+        return self._handle_login_response(response.status_code, jsondata, response.cookies)
 
-        self.myaussie_cookie = response.cookies.get('myaussie_cookie')
-        if self.myaussie_cookie:
-            logger.debug(f"Login Cookie: {self.myaussie_cookie}")
-        return True
 
-    def _has_token_expired(self):
-        """ Returns bool of if the token has expired """
-        if time() > self.token_expires:
-            return True
-        return False
+    def do_login_check(self, skip_login_check: bool) -> None:
+        """ checks if we're skipping the login check and logs in if necessary """
+        if not skip_login_check:
+            self.logger.debug("skip_login_check false")
+            if self._has_token_expired():
+                self.logger.debug("token has expired, logging in...")
+                self.login()
 
     def request_get(self, skip_login_check: bool = False, **kwargs):
         """ Performs a GET request and logs in first if needed.
 
         Returns the `requests.Response` object."""
-        if not skip_login_check:
-            logger.debug("skip_login_check false")
-            if self._has_token_expired():
-                logger.debug("token has expired, logging in...")
-                self.login()
+        self.do_login_check(skip_login_check)
         if 'cookies' not in kwargs:
             kwargs['cookies'] = {'myaussie_cookie' : self.myaussie_cookie}
         response = self.session.get(**kwargs)
@@ -105,24 +89,16 @@ class AussieBB():
 
         Returns a dict of the JSON response.
         """
-        if not skip_login_check:
-            logger.debug("skip_login_check false")
-            if self._has_token_expired():
-                logger.debug("token has expired, logging in...")
-                self.login()
         if 'cookies' not in kwargs:
             kwargs['cookies'] = {'myaussie_cookie' : self.myaussie_cookie}
+        self.do_login_check(skip_login_check)
         response = self.session.get(**kwargs)
         response.raise_for_status()
         return response.json()
 
     def request_post(self, skip_login_check: bool = False, **kwargs):
         """ Performs a POST request and logs in first if needed."""
-        if not skip_login_check:
-            logger.debug("skip_login_check false")
-            if self._has_token_expired():
-                logger.debug("token has expired, logging in...")
-                self.login()
+        self.do_login_check(skip_login_check)
         if 'cookies' not in kwargs:
             kwargs['cookies'] = {'myaussie_cookie' : self.myaussie_cookie}
         if 'headers' not in kwargs:
@@ -135,8 +111,7 @@ class AussieBB():
         """ Grabs the customer details.
 
         Returns a dict """
-        frame = inspect.currentframe()
-        url = get_url(inspect.getframeinfo(frame).function)
+        url = self.get_url("get_customer_details")
         querystring = {"v":"2"}
         responsedata = self.request_get_json(url=url,
                                     params=querystring,
@@ -160,7 +135,12 @@ class AussieBB():
         return False
 
 
-    def get_services(self, page: int = 1, servicetypes: list=None, use_cached: bool=False):
+    def get_services(
+        self,
+        page: int = 1,
+        servicetypes: list=None,
+        use_cached: bool=False,
+        ) -> Optional[List[Dict[str, Any]]]:
         """ Returns a `list` of `dicts` of services associated with the account.
 
             If you want a specific kind of service, or services,
@@ -169,38 +149,44 @@ class AussieBB():
             If you want to use cached data, call it with `use_cached=True`
         """
         if use_cached:
-            logger.debug("Using cached data for get_services.")
+            self.logger.debug("Using cached data for get_services.")
             self._check_reload_cached_services()
         else:
-            frame = inspect.currentframe()
-            url = get_url(inspect.getframeinfo(frame).function)
+            url = self.get_url("get_services")
             querystring = {'page' : page}
 
             responsedata = self.request_get_json(url=url, params=querystring)
             # cache the data
-            self.services_last_update = time()
-            self.services = responsedata.get('data')
+            self.services_last_update = int(time())
+            if "data" not in responsedata:
+                raise ValueError("Couldn't get 'data' element from response.")
+            self.services = responsedata["data"]
+
+        # TODO: validate the expected fields in the service (type, name, plan, description, service_id at a minimum)
 
         # only filter if we need to
-        if servicetypes:
-            logger.debug("Filtering services based on provided list: {}", servicetypes)
-            filtered_responsedata = []
-            for service in self.services:
-                if service.get('type') in servicetypes:
-                    filtered_responsedata.append(service)
-                else:
-                    logger.debug("Skipping as type=={} - {}", service.get('type'), service)
-            return filtered_responsedata
+        if servicetypes is not None:
+            self.logger.debug("Filtering services based on provided list: %s", servicetypes)
+            filtered_responsedata: List[str] = []
+            if self.services is not None:
+                for service in self.services:
+                    if "type" not in service:
+                        raise ValueError(f"No type field in service info: {service}")
+                    if service["type"] in servicetypes:
+                        filtered_responsedata.append(service)
+                    else:
+                        self.logger.debug("Skipping as type==%s - %s", service["type"], service)
+                return filtered_responsedata
 
         return self.services
 
 
-    def account_transactions(self):
+    def account_transactions(self) -> Dict[str, AccountTransaction]:
         """ Pulls the data for transactions on your account.
 
             Returns a dict where the key is the month and year of the transaction.
 
-            Keys: `['current', 'pending', 'available', 'filters', 'typicalEveningSpeeds']`
+            Keys: `['id', 'type', 'time', 'description', 'amountCents', 'runningBalanceCents']`
 
             Example output:
 
@@ -217,8 +203,7 @@ class AussieBB():
             ],
             ```
             """
-        frame = inspect.currentframe()
-        url = get_url(inspect.getframeinfo(frame).function)
+        url = self.get_url("account_transactions")
         return self.request_get_json(url=url)
 
 
@@ -227,35 +212,36 @@ class AussieBB():
 
             This returns the bare response object, parsing the result is an exercise for the consumer. It's a PDF file.
         """
-        frame = inspect.currentframe()
-        url = get_url(inspect.getframeinfo(frame).function, {'invoice_id' : invoice_id})
+        url = self.get_url("billing_invoice", {'invoice_id' : invoice_id})
         return self.request_get_json(url=url)
 
     def account_paymentplans(self):
         """ Returns a dict of payment plans for an account """
-        frame = inspect.currentframe()
-        url = get_url(inspect.getframeinfo(frame).function)
+        url = self.get_url("account_paymentplans")
         return self.request_get_json(url=url)
 
-    def get_usage(self, service_id: int):
+    def get_usage(self, service_id: int, use_cached: bool=True) -> Dict[str, Any]:
         """
         Returns a dict of usage for a service.
 
         If it's a telephony service (`type=PhoneMobile`) it'll pull from the telephony endpoint.
 
         """
+        if self.services is None:
+            self.get_services(use_cached=use_cached)
 
-        if not self.services:
-            self.get_services(use_cached=True)
-        for service in self.services:
-            if service_id == service.get('service_id'):
-                if service.get('type') in ['PhoneMobile']:
-                    return self.telephony_usage(service_id)
-        frame = inspect.currentframe()
-        url = get_url(inspect.getframeinfo(frame).function, {'service_id' : service_id})
-        return self.request_get_json(url=url)
+        if self.services is not None:
+            for service in self.services:
+                if service_id == service["service_id"]:
+                    # throw an error if we're trying to parse something we can't
+                    self.validate_service_type(service)
+                    if service["type"] in PHONE_TYPES:
+                        return self.telephony_usage(service_id)
+            url = self.get_url("get_usage", {'service_id' : service_id})
+            return self.request_get_json(url=url)
+        return {}
 
-    def get_service_tests(self, service_id: int):
+    def get_service_tests(self, service_id: int) -> List[ServiceTest]:
         """ Gets the available tests for a given service ID
         Returns list of dicts
 
@@ -271,24 +257,22 @@ class AussieBB():
 
         This has a habit of throwing 400 errors if you query a VOIP service...
         """
-        frame = inspect.currentframe()
-        url = get_url(inspect.getframeinfo(frame).function, {'service_id' : service_id})
-        return self.request_get_json(url=url)
+        url = self.get_url("get_service_tests", {'service_id' : service_id})
+        results: List[ServiceTest] = self.request_get_json(url=url)
+        return results
 
     def get_test_history(self, service_id: int):
         """ Gets the available tests for a given service ID
 
         Returns a list of dicts with tests which have been run
         """
-        frame = inspect.currentframe()
-        url = get_url(inspect.getframeinfo(frame).function, {'service_id' : service_id})
+        url = self.get_url("get_test_history", {'service_id' : service_id})
         return self.request_get_json(url=url)
 
     def test_line_state(self, service_id: int):
         """ Tests the line state for a given service ID """
-        frame = inspect.currentframe()
-        url = get_url(inspect.getframeinfo(frame).function, {'service_id' : service_id})
-        logger.debug("Testing line state, can take a few seconds...")
+        url = self.get_url("test_line_state", {'service_id' : service_id})
+        self.logger.debug("Testing line state, can take a few seconds...")
         response = self.request_post(url=url)
         return response.json()
 
@@ -308,10 +292,10 @@ class AussieBB():
         if not test_links:
             return False
         if len(test_links) != 1:
-            logger.debug(f"Too many tests? {test_links}")
+            self.logger.debug("Too many tests? %s", test_links)
 
-        test_name = test_links[0].get('name')
-        logger.debug(f"Running {test_name}")
+        test_name = test_links[0]["name"]
+        self.logger.debug("Running %s", test_name)
         if test_method == 'get':
             return self.request_get_json(url=test_links[0].get('link'))
         return self.request_post(url=test_links[0].get('link')).json()
@@ -322,8 +306,7 @@ class AussieBB():
             Keys: `['current', 'pending', 'available', 'filters', 'typicalEveningSpeeds']`
 
         """
-        frame = inspect.currentframe()
-        url = get_url(inspect.getframeinfo(frame).function, {'service_id' : service_id})
+        url = self.get_url("service_plans", {'service_id' : service_id})
         return self.request_get_json(url=url)
 
     def service_outages(self, service_id: int):
@@ -349,8 +332,7 @@ class AussieBB():
             }
             ```
         """
-        frame = inspect.currentframe()
-        url = get_url(inspect.getframeinfo(frame).function, {'service_id' : service_id})
+        url = self.get_url("service_outages", {'service_id' : service_id})
         return self.request_get_json(url=url)
 
     def service_boltons(self, service_id: int):
@@ -371,8 +353,7 @@ class AussieBB():
             }]
             ```
             """
-        frame = inspect.currentframe()
-        url = get_url(inspect.getframeinfo(frame).function, {'service_id' : service_id})
+        url = self.get_url("service_boltons", {'service_id' : service_id})
         return self.request_get_json(url=url)
 
     def service_datablocks(self, service_id: int):
@@ -389,8 +370,7 @@ class AussieBB():
             }
             ```
         """
-        frame = inspect.currentframe()
-        url = get_url(inspect.getframeinfo(frame).function, {'service_id' : service_id})
+        url = self.get_url("service_datablocks", {'service_id' : service_id})
         return self.request_get_json(url=url)
 
     def telephony_usage(self, service_id: int):
@@ -404,8 +384,7 @@ class AussieBB():
             {"national":{"calls":0,"cost":0},"mobile":{"calls":0,"cost":0},"international":{"calls":0,"cost":0},"sms":{"calls":0,"cost":0},"internet":{"kbytes":0,"cost":0},"voicemail":{"calls":0,"cost":0},"other":{"calls":0,"cost":0},"daysTotal":31,"daysRemaining":2,"historical":[]}
             ```
             """
-        frame = inspect.currentframe()
-        url = get_url(inspect.getframeinfo(frame).function, {'service_id' : service_id})
+        url = self.get_url("telephony_usage", {'service_id' : service_id})
         return self.request_get_json(url=url)
 
     def support_tickets(self):
@@ -414,8 +393,7 @@ class AussieBB():
             Dict keys: `['ref', 'create', 'updated', 'service_id', 'type', 'subject', 'status', 'closed', 'awaiting_customer_reply', 'expected_response_minutes']`
 
             """
-        frame = inspect.currentframe()
-        url = get_url(inspect.getframeinfo(frame).function)
+        url = self.get_url("support_tickets")
         return self.request_get_json(url=url)
 
     def get_appointment(self, ticketid: int):
@@ -423,8 +401,7 @@ class AussieBB():
 
             Dict keys: `['ref', 'create', 'updated', 'service_id', 'type', 'subject', 'status', 'closed', 'awaiting_customer_reply', 'expected_response_minutes']`
             """
-        frame = inspect.currentframe()
-        url = get_url(inspect.getframeinfo(frame).function, {'ticketid' : ticketid})
+        url = self.get_url("get_appointment", {'ticketid' : ticketid})
         return self.request_get_json(url=url)
 
 
@@ -433,6 +410,5 @@ class AussieBB():
 
             Dict keys: `['id', 'first_name', 'last_name', 'email', 'dog', 'home_phone', 'work_phone', 'mobile_phone', 'work_mobile', 'primary_contact']`
             """
-        frame = inspect.currentframe()
-        url = get_url(inspect.getframeinfo(frame).function)
+        url = self.get_url("account_contacts")
         return self.request_get_json(url=url)
